import logging
from typing import List, Union, Tuple, Optional

import numpy as np
import treefiles as tf
from treefiles.baseio_.param import r


class Feature(tf.BaseIO):
    def __init__(
        self,
        *args,
        ref=None,  # : tf.TValue
        error: tf.TValue = None,
        weight: tf.TValue = 1,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.ref = ref
        self.weight = weight
        self._error = error

        self.registered.add("ref")
        self.registered.add("error")
        self.registered.add("weight")

    @property
    def error(self):
        if self._error is None:
            if self.ref is None:
                return
            if self.value is not None:
                return self.error2
        else:
            return self._error

    @error.setter
    def error(self, value: tf.TValue):
        self._error = value

    @property
    def error2(self) -> Optional[tf.TValue]:
        """
        Return (x - ref) ** 2
        """
        if self.value is None or self.ref is None:
            return
        if isinstance(self.ref, str):
            self.ref = eval(self.ref)
        try:
            return np.square(self.value - self.ref)
        except TypeError:
            return

    @property
    def error_rel(self) -> Optional[tf.TValue]:
        """
        Return relative error (x - ref) / ref
        If ref=0, returns self.error2
        """
        if self.value is None or self.ref is None:
            return
        if isinstance(self.ref, float) and self.ref == 0:
            return self.error2
        try:
            return (self.value - self.ref) / np.abs(self.ref)
        except TypeError:
            return

    @property
    def table(self):
        return Features(self).table


# class Features(tf.Bases[str, Feature]):  # >= python 3.9
class Features(tf.Bases):
    inner_class = Feature

    def build_table(self) -> Tuple[tf.Bases.THeader, tf.Bases.TData]:
        # fmt: off
        header = ["Name", "Value", "Ref", "Error", "Error L2", "Error rel", "Weight", "Unit"]
        data = [
            [x.name, r(x.value), r(x.ref), r(x.error), r(x.error2), r(x.error_rel), x.weight, x.unit] for x in self.values()
        ]
        # fmt: on
        return header, data

    def set_rel_error(self):
        for x in self.values():
            x.error = abs(x.error_rel) * 100
        return self

    def set_error_L2(self):
        for x in self.values():
            x.error = x.error2
        return self

    @property
    def errors(self) -> List[str]:
        return [x.error for x in self.values()]


TFeatures = Union[List[Feature], Features]

log = logging.getLogger(__name__)
