import logging
import sys
from typing import List, Callable

import matplotlib.pyplot as plt
import numpy as np
import treefiles as tf
from matplotlib import animation
from matplotlib.collections import PathCollection
from scipy import stats, interpolate
from tqdm import tqdm

from gcma.io.utils import colors
from gcma.plotter.confid_ell import confidence_ellipse


def _plot_dist(self, save_fname: str = None, show: bool = False, corner: bool = False):
    """
    Generates a pair plot animation

    Parameters
    ----------
    save_fname: animation file name
    show: call plt.show
    corner: if 3d surface interpolation should be plotted: only relevant with 2 parameters
    """
    plt.style.use("ggplot")
    # plt.axis("equal")

    if save_fname is None:
        save_fname = self.out_dir.path("generations.mp4")

    params = self.parent.rescale(self.parent.param_history)
    n_iter, n_pop, n_params = params.shape

    sf = n_params * 2
    fig, axs = plt.subplots(
        nrows=n_params, ncols=n_params, figsize=(sf, sf), sharex="col", sharey="row"
    )
    grid = GridPlot(
        params, fig, axs, n_iter, energy=self.parent.energy_history, corner=corner
    )
    grid.init_axes(self.parent.params)
    grid.init()

    ani = animation.FuncAnimation(
        fig, grid.animate, frames=n_iter, interval=300, blit=True
    )
    ani.save(save_fname)

    if show:
        plt.show()
    plt.close(fig)
    plt.style.use("default")


class GridPlot(list):
    def __init__(
        self, params, fig, axs, n_iter, energy, corner: bool, pbar: bool = True
    ):
        self.params = params
        self.fig = fig
        self.axs = axs
        self.n_iter = n_iter
        self.n_params = params.shape[2]
        self.energy = np.array(energy)
        self._pbar = tqdm(total=n_iter, disable=not pbar, file=sys.stdout)
        self.names = None

        super().__init__(
            [
                DistProj(self, Iterators.tri_inf),
                KDE(self, Iterators.diag),
            ]
        )
        trisup = Plot3D if corner else EmptySubPlot
        self.append(trisup(self, Iterators.tri_sup))

    def __del__(self):
        self._pbar.close()

    def init_axes(self, params: tf.TParams):
        self.names = []
        plt.rc("axes", labelsize=8)
        for i, x in enumerate(params.values()):
            name = x.get_pname()
            self.names.append(name)
            self.axs[-1, i].set_xlim(x.bounds)
            self.axs[-1, i].set_xlabel(name)
            self.axs[i, 0].set_ylim(x.bounds)
            self.axs[i, 0].set_ylabel(name)
            self.axs[i, i].axes.get_yaxis().set_visible(False)

    def init(self) -> List[PathCollection]:
        actors = []
        for x in self:
            actors.extend(x.init())
        plt.tight_layout()
        return actors

    def animate(self, k):
        self._pbar.update()

        actors = []
        for x in self:
            acts = x.plot(
                self.params[k, ...],
                it=k,
                n_iter=self.n_iter,
            )
            actors.extend(acts)
        return actors


class SubPlot:
    def __init__(self, parent: GridPlot, iterator: Callable):
        self.parent = parent
        self.n = parent.n_params
        self.actors = np.array([[{} for _ in range(self.n)] for _ in range(self.n)])
        self._iterator = iterator

    def init(self) -> List[PathCollection]:
        raise NotImplementedError

    def plot(self, params: np.ndarray, **kwargs) -> List[PathCollection]:
        return []

    def inds(self):
        for i, j in self._iterator(self.n):
            yield i, j

    def ax_ac(self):
        for i, j in self.inds():
            yield i, j, self.parent.axs[i, j], self.actors[i, j]

    @property
    def all_actors(self):
        actors = []
        for x in self.actors.flatten():
            actors.extend(x.values())
        return actors


class DistProj(SubPlot):
    def init(self) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            ac.update(
                dict(
                    trace=ax.plot([], [], colors["grey"], lw=1, alpha=0.5)[0],
                    scat=ax.scatter([], [], color=colors["blue"]),
                    mu=ax.scatter([], [], color=colors["red"], marker="+", s=50),
                )
            )
        return self.all_actors

    def plot(self, params: np.ndarray, **kwargs) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            x, y = params[:, j], params[:, i]
            values = np.stack((x, y), axis=1)
            xc, yc = x.mean(), y.mean()

            trace = ac["trace"]
            trace.set_xdata(np.append(trace.get_xdata(), xc))
            trace.set_ydata(np.append(trace.get_ydata(), yc))

            ac["mu"].set_offsets([xc, yc])
            ac["scat"].set_offsets(values)

            ax.patches = []
            confidence_ellipse(x, y, ax, edgecolor=colors["red"], linewidth=1.5)

        return self.all_actors


class KDE(SubPlot):
    def init(self) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            ac.update({"kde": ax.plot([], [], color=colors["blue"], lw=1.5)[0]})

            if i == j == 0:
                text = ax.text(
                    0.2,
                    1,
                    "",
                    bbox={"facecolor": "w", "alpha": 0.5, "pad": 5},
                    transform=ax.transAxes,
                    ha="center",
                )
                ac.update({"title": text})
        return self.all_actors

    def plot(self, params: np.ndarray, **kwargs) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            data = params[:, i]

            lim = ax.get_ylim()
            x = np.linspace(*lim, 500)
            y = stats.gaussian_kde(data)(x)
            m = 1 if abs(np.max(y)) < 1e-10 else np.max(y)
            y = y / 1.1 / m * (lim[1] - lim[0]) + lim[0]

            ac["kde"].set_xdata(x)
            ac["kde"].set_ydata(y)

            ax.collections.clear()
            ax.fill_between(x, lim[0], y, facecolor=colors["blue"], alpha=0.3)

            if i == j == 0:
                n_iter = kwargs.get("n_iter")
                it = kwargs.get("it")
                ac["title"].set_text(f"Gen {it+1}/{n_iter}")

        return self.all_actors


class EmptySubPlot(SubPlot):
    def init(self) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            self.parent.fig.delaxes(ax)
        return []


class Plot3D(SubPlot):
    def init(self) -> List[PathCollection]:
        n = self.n
        for i, j, ax, ac in self.ax_ac():
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            ax.remove()
            new_ax = self.parent.fig.add_subplot(n, n, i * n + j + 1, projection="3d")
            self.parent.axs[i, j] = new_ax

            # extrapolate surface based on all results
            p = self.parent.params
            ni, ps, pa = p.shape
            p = p.reshape((ni * ps, pa))
            x, y = p[..., i], p[..., j]
            points = np.stack((x, y), axis=1)

            values = self.parent.energy.flatten()

            # ensure no error in stacks
            try:
                assert self.parent.params[2, 5, i] == points[2 * ps + 5, 0]
                assert self.parent.energy[2, 5] == values[2 * ps + 5]
            except IndexError:
                pass

            grid_x = np.linspace(*ax.get_xlim(), 100)
            grid_y = np.linspace(*ax.get_ylim(), 100)
            xd, yd = np.meshgrid(grid_x, grid_y)
            zd = interpolate.griddata(points, values, (xd, yd), method="cubic")

            new_ax.plot_surface(
                xd,
                yd,
                zd,
                rstride=1,
                antialiased=False,
                color=colors["grey"],
                alpha=0.1,
            )
            new_ax.set_facecolor("w")
            new_ax.set_xlim(*ylim)
            new_ax.set_xlabel(self.parent.names[i])
            new_ax.set_ylim(*xlim)
            new_ax.set_ylabel(self.parent.names[j])
            new_ax.axes.get_zaxis().set_visible(True)

            ac.update({"scat": new_ax.scatter([], [], [], color=colors["blue"])})

        return self.all_actors

    def plot(self, params: np.ndarray, **kwargs) -> List[PathCollection]:
        for i, j, ax, ac in self.ax_ac():
            x, y = params[:, i], params[:, j]
            z = self.parent.energy[kwargs.get("it"), :]

            ac["scat"]._offsets3d = (x, y, z)

        return self.all_actors


class Iterators:
    @staticmethod
    def tri_sup(n):
        for i in range(1, n):
            for j in range(i):
                yield j, i

    @staticmethod
    def tri_inf(n):
        for i in range(1, n):
            for j in range(i):
                yield i, j

    @staticmethod
    def diag(n):
        for i in range(n):
            yield i, i


log = logging.getLogger(__name__)
