import logging
from typing import List, Tuple

import numpy as np
import treefiles as tf


class Prior:
    def __init__(
        self,
        mean: tf.TValue = None,
        cov: np.ndarray = None,
        gamma: float = 1,
        do_update: bool = True,
    ):
        """
        :param gamma: weigth of the regularisation term
        """
        self.mean = None if mean is None else np.array(mean)
        self.cov = None if cov is None else np.array(cov)
        self.gamma = gamma
        self.do_update = do_update
        self.slice = np.s_[:]
        self.rules_ = None

    def get_distance(self, params: tf.TParams) -> float:
        """
        Get the distance between current parameters and prior knowledge

        :param params: non normalised params of one sample
        """
        #
        # Maybe use `scipy.spatial.distance.cdist` :
        # cdist(x, np.zeros(3)[np.newaxis, :], metric="mahalanobis")
        #
        if self.mean is None or self.cov is None:
            return 0

        vals = np.array(params.vals)[self.slice]
        no = (vals - self.mean)[:, np.newaxis]
        d = (no.T @ np.linalg.inv(self.cov) @ no).squeeze()
        return d * self.gamma

    def update(self, results: np.ndarray, force: bool = False):
        """
        Updates prior based on database parameters distribution

        :param results: parameters value for each member of the database, size (n_patients, n_params)
        :param force:
        """
        if self.do_update or force:
            self.mean = results.mean(axis=0)

            # full matrix update
            # self.cov = np.cov(results.T)

            # diagonal matrix update
            res = np.cov(results.T)
            if res.ndim > 1:
                res = np.diag(np.diag(res))
            self.cov = res

    @classmethod
    def from_rules(
        cls,
        rules: List[Tuple[str, float, float]],
        params: tf.TParams,
        pop_size: int,
        **kwargs,
    ):
        names = params.names
        idx, res = [], np.zeros((pop_size, len(rules)))
        for j, (p, mean, std) in enumerate(rules):
            idx.append(names.index(p))
            res[:, j] = np.random.normal(mean, std, pop_size)

        c = cls(**kwargs)
        c.rules_ = rules
        c.update(res, force=True)
        c.slice = np.r_[idx]
        return c

    def __repr__(self):
        s = f"Prior rules: {self.rules_} (gamma={self.gamma})"
        return s


log = logging.getLogger(__name__)

# if __name__ == "__main__":
#     from gcma import Params, Param
#
#     pa = Params(Param("x1", 1), Param("x2", 1), Param("x3", 10))
#     rules = [("x1", 1, 0.3), ("x3", 2, 0.3)]
#     prior = Prior.from_rules(rules, pa, 500, gamma=1e-2)
#
#     print(prior.get_distance(pa))
#     print(prior.mean)
#     print(prior.cov)
