import logging
import pickle
import time
from datetime import datetime
from typing import List, Tuple, Union

from treefiles.baseio_.baseio_ import IOEncoder

try:
    import cma
except ImportError as e:
    print("You must install cma: 'pip install cma'")
    raise e

import numpy as np
import treefiles as tf
from cma.evolution_strategy import CMAEvolutionStrategyResult, CMAEvolutionStrategy
from gcma.constraints.constraint import TConstraints, Constraints
from gcma.io import *
from gcma.prior.prior import Prior


WRAPPER_TREE_OUT = tf.curDirs(__file__, "out.tree")


class CMA:
    # Main cmaes output tree defined here (filenames and directories structure)
    __tree_file__ = WRAPPER_TREE_OUT

    def __init__(
        self,
        params: tf.TParams,
        features: TFeatures,
        options: Options,
        output_dir: Union[tf.Tree, str],
        sigma0: float = 2,
        scale_factor: float = 10,
        result_type: str = "xbest",
        termination_options: TermOptions = None,
        constraints: TConstraints = None,
        prior: Prior = None,
        force_no_stdout: bool = False,
        default_params: tf.TParams = None,
        **kwargs,
    ):
        """
        Wraps the pycma implementation of N.Hansen: https://github.com/CMA-ES/pycma.
        Additional features such as constraints, prior knowledge, plots, animations, ...

        :param params: parameters
        :param features: features
        :param options: options passed to pycma
        :param output_dir: path of directory for exports
        :param sigma0: initial parameters deviation
        :param scale_factor: used in the scaling (between 0 and scale_factor)
        :param result_type: attribute name for result pycma object
        :param termination_options: options for the termination callback
        :param constraints: list of parameters constraints
        :param prior: gcma.Prior instance, prior knowledge
        :param force_no_stdout: if you do not want pycma to write to sys.stdout
        :param kwargs: additional ptions passed to pycma
        """
        self.output_dir = self._init_output_dir(output_dir).dump()

        self.params = tf.Params(params)
        # New: all parameters defined with a None `bounds` attribute will be considered
        # as default parameters
        self.def_params = tf.Params([x for x in self.params.elems if x.bounds is None])
        if default_params is not None:
            self.def_params += default_params
        self.params = tf.Params([x for x in self.params.elems if x.bounds is not None])

        self.features = Features(features)
        self.dump_legends()

        self.sigma0 = sigma0
        self.scale_factor = scale_factor
        self.result_type = result_type
        self.termination_options = (
            TermOptions() if termination_options is None else termination_options
        )
        self.constraints = Constraints(constraints)
        self.prior = Prior() if prior is None else prior
        self.force_no_stdout = force_no_stdout
        self._finished = False
        self.time_start_ = None
        self.time_end_ = None

        # fmt: off
        self.param_history = []  # parameters values associated to each sample
        self.features_history = []  # values of the features associated to the best sample
        self.energy_history_ft = []  # errors of the features associated to sample
        self.energy_history = []  # global energy associated to each sample (score + constraint + prior)
        self.energy_history_score = []  # score value associated to each sample
        self.energy_history_constraints = []  # constraint value associated to each sample
        self.energy_history_prior = []  # regularization (prior) value associated to each sample
        # fmt: on
        self.history = History()

        self.es: cma.CMAEvolutionStrategy = None
        options.update(kwargs)
        self._init(**dict(options))

        self.plotter = None
        self._init_plotter()

    @classmethod
    def from_previous(cls, path: tf.TS):
        """
        Used to load an instance of CMA based on previous exports

        :param path: path to directory containing previous exports
        """
        out = cls._init_output_dir(path)
        params, features, default_params = load_legends(out.legends)
        options = tf.load_json(out.options)
        c = cls(params, features, options, path, default_params=default_params)
        c.init_previous()
        return c

    def init_previous(self):
        """
        Initialize new CMA instance
        """
        with open(self.output_dir.cma_pickle, "rb") as f:
            self.es = pickle.loads(f.read())
        self.es.stop().clear()

        self.param_history = np.load(self.output_dir.pa).tolist()
        self.energy_history = np.load(self.output_dir.en).tolist()
        self.energy_history_ft = np.load(self.output_dir.enf).tolist()
        self.energy_history_score = np.load(self.output_dir.score).tolist()
        self.energy_history_constraints = np.load(self.output_dir.cts).tolist()
        self.energy_history_prior = np.load(self.output_dir.prior).tolist()
        self.features_history = np.load(self.output_dir.fv).tolist()

        self.termination_options = TermOptions.from_dict(
            tf.load_json(self.output_dir.term)
        )
        self.termination_options["term_on_date"] = False

        self.history = History.load(self.output_dir.history)

    def _init_plotter(self):
        from gcma.plotter.plotter import Plotter

        self.plotter = Plotter(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Calls the _finish method
        """
        if not self._finished:
            self._finish()

    @property
    def n_params(self):
        """
        Number of parameters
        """
        return len(self.params)

    @property
    def n_features(self):
        """
        Number of features
        """
        return len(self.features)

    @property
    def popsize(self):
        """
        Size of the population
        """
        return self.es.opts["popsize"]

    @property
    def n_iters(self):
        """
        Maximum number of iterations
        """
        return self.es.opts["maxiter"]

    @classmethod
    def _init_output_dir(cls, tree) -> tf.T:
        """
        Initialize filenames
        :param tree: parent directory
        """
        tree_ = tf.Tree.from_file(cls.__tree_file__)
        tree_.root = tree
        return tree_

    @property
    def bounds(self) -> np.ndarray:
        """
        Parameters bounds
        """
        return np.array([x.bounds for x in self.params.values()])

    def _backup(self):
        """
        Called at the end of each generation, saves current state to history
        """
        best_i = np.argmin(self.energy_history[-1])

        # Best set of parameters
        # equivalent to `self.es.best.last.x`
        bestx = self.param_history[-1][best_i, ...]
        params = self.params.assign(self.rescale(bestx))
        if not np.array_equal(bestx, self.es.best.last.x):
            log.error(f"gcma 'wrapper' failed")
            log.debug(bestx)
            log.debug(self.es.best.last.x)

        # Features values associated to best parameters
        bestfv = self.features_history[-1][best_i, ...]
        features = self.features.assign(bestfv)

        # Features errors associated to best parameters
        bestfe = self.energy_history_ft[-1][best_i, ...]
        features = features.assign(bestfe, "error")
        assert np.isclose(features.errors, bestfe).all()

        # Global energy associated to best parameters
        # equivalent to `self.es.best.last.f`
        bestf = self.energy_history[-1][best_i]
        if not np.array_equal(bestf, self.es.best.last.f):
            log.error(f"gcma 'wrapper' failed")
            log.debug(bestf)
            log.debug(self.es.best.last.f)

        self.history.add(params=params, features=features, fvalue=bestf)

    def get_history(self, t: str):
        """
        Return the history, ie the best sample among the population at generation `t`
        """
        if len(self.history) == 0:
            return "No history to display"
        return getattr(self.history, t)

    @property
    def last(self):
        """
        Get the history of last generation
        """
        return self.get_history("last")

    @property
    def best(self):
        """
        Get the history of the best generation
        """
        return self.get_history("best")

    def _init(self, **options):
        """
        Initialize cma.CMAEvolutionStrategy
        """
        initial_values = np.array([x.initial_value for x in self.params.values()])
        n_val = len(self.bounds)

        if initial_values is None:
            initial_values = [self.scale_factor / 2] * n_val
        else:
            initial_values = self.descale(initial_values)

        inopts = {
            "bounds": ([0] * n_val, [self.scale_factor] * n_val),
            "verb_filenameprefix": f"{self.output_dir.cma_out.abs()}/",
            "termination_callback": self.termination_callback,
        }
        options.update(inopts)

        with tf.no_stdout(self.force_no_stdout):
            self.es = cma.CMAEvolutionStrategy(
                x0=initial_values, sigma0=self.sigma0, inopts=options
            )

        self.dump_options()

    def termination_callback(self, es: CMAEvolutionStrategy) -> bool:
        """
        Custom termination callback

        :param es: the current CMAEvolutionStrategy instance
        :return: True if a termination condition is met, False otherwise
        """
        opts = self.termination_options

        is_over = opts.is_over()
        is_tol = False
        if es.countiter > 1:
            values = [x.error for x in self.history.best.features.values()]
            is_tol = opts.is_tol(values)

        return is_over or is_tol

    def dump_legends(self):
        """
        Helper to export parameters and features
        """
        tf.dump_json(
            self.output_dir.legends,
            {
                "params": self.params,
                "features": self.features,
                "default_params": self.def_params,
            },
            cls=IOEncoder,
        )

    def load_legends(self) -> Tuple[tf.TParams, TFeatures]:
        """
        Helper to load parameters and features
        """
        return load_legends(self.output_dir.legends)

    def dump_options(self):
        """
        Helper to export options defined by `Options.to_save` method
        """
        options = {k: self.es.opts[k] for k in Options.to_save()}
        tf.dump_json(self.output_dir.options, options)

    def rescale(self, x):
        """
        From normalized CMA-ES values (in [0, scale_factor]) to real world

        :param x: normalized parameters values, between 0 and self.scale_factor
        :return: non normalized parameters values
        """
        b, s = self.bounds, self.scale_factor
        return x * (b[:, 1] - b[:, 0]) / s + b[:, 0]

    def descale(self, x):
        """
        Normalize from real world coords to CMAES world (btw 0 and scale_factor)

        :param x: non normalized parameters values
        :return: normalized parameters values, between 0 and self.scale_factor
        """
        b, s = self.bounds, self.scale_factor
        return (x - b[:, 0]) * s / (b[:, 1] - b[:, 0])

    def energy(
        self, params: List[tf.TParams], **kwargs
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Main method to compute the errors related to each sample features

        :param params: the set of parameters for each sample of the population (list of size `popsize`)
        :return: a tuple of 2 arrays of size (popsize, n_features), the first one with the features error,
            the second with features values
        """
        raise NotImplementedError
        # return errors, values

    def energy_sum(self, features_error: np.ndarray) -> np.ndarray:
        """
        Main method to concatenate the features errors for each sample

        :param features_error: an array of size (popsize, n_features)
        :return: an array of size (popsize,), where each row contains the global sample features error
        """
        assert features_error.shape == (self.popsize, self.n_features)
        features_error[np.where(np.isnan(features_error))] = 1e10
        return np.sum(features_error, axis=1)

    def energy_constraints(self, params: List[tf.TParams]) -> np.ndarray:
        """
        Main method to concatenate the constraints contributions for each sample

        :param params: the set of parameters for each sample of the population (list of size `popsize`)
        :return: an array of size (popsize,), where each row contains the sum of all constraints
        """
        cts = np.zeros(self.popsize)

        for i, p in enumerate(params):
            contribs = self.constraints.apply(p)
            contribs = np.array([x[1] for x in contribs])
            cts[i] = contribs.sum()

        return cts

    def _generation(self):
        """
        Implements a CMA-ES generation
        """
        # log.info(f"CMAES iteration={self.es.countiter}")

        values_to_test = np.array(self.es.ask())
        self.param_history.append(values_to_test)
        scaled = self.rescale(values_to_test)

        # list of size `self.popsize`. Each sublist contains the parameters
        params = [self.params.assign(scaled[i, :]) for i in range(self.popsize)]

        features_error, features_value = self.energy(params, n_iter=self.es.countiter)
        self.energy_history_ft.append(features_error)
        self.features_history.append(features_value)

        samples_error = self.energy_sum(features_error)
        self.energy_history_score.append(samples_error)

        contraints = self.energy_constraints(params)
        self.energy_history_constraints.append(contraints)

        prior_regul = self.energy_prior(params)
        self.energy_history_prior.append(prior_regul)

        energy = samples_error + contraints + prior_regul
        self.energy_history.append(energy)

        self.es.tell(values_to_test, energy)
        self.es.logger.add()

        with tf.no_stdout(self.force_no_stdout):
            self.es.disp()

        self._backup()
        self.end_generation()

    def energy_prior(self, params: List[tf.TParams]) -> np.ndarray:
        """
        Main method to concatenate the prior contributions for each sample

        :param params: the set of parameters for each sample of the population (list of size `popsize`)
        :return: an array of size (popsize,), where each row contains the value of the prior regularisation
        """
        return np.array([self.prior.get_distance(x) for x in params])

    def run(self, *exceptions) -> CMAEvolutionStrategyResult:
        """
        Start the CMA-ES algorithm

        :param exceptions: list of exceptions
        :return: a CMAEvolutionStrategyResult object
        """
        self._finished = False
        log.info(f"Starting {type(self)} to file://{self.output_dir.abs()}")
        self.time_start_ = time.perf_counter()

        while not self.es.stop():
            if len(exceptions) == 0:
                self._generation()
            else:
                try:
                    self._generation()
                except exceptions as e:
                    log.critical(f"Exiting CMA-ES main loop: {e}")
                    break

        self._finish()
        return self.es.result

    def test_energy(self, list_params, n_iter, **_):
        """
        The feature value (same for all features) is the sum of the
        parameters, plus a gaussian noise
        The feature error (same for all features) is `abs(feature value)`
        """
        params_ = np.array([x.vals for x in list_params])
        sum_params = np.sum(params_, axis=1)[:, np.newaxis]
        value_features = np.repeat(sum_params, self.n_features, axis=1)
        vals = value_features + np.random.normal(
            10 - n_iter,
            0.5 * np.mean(np.abs(sum_params)),
            (len(list_params), self.n_features),
        )
        # print(np.abs(vals), vals)
        return np.abs(vals), vals

    def test_run(self):
        log.debug("Starting test_run on CMA instance")
        self.energy = self.test_energy
        return self.run()

    def end_generation(self):
        """
        Called at the end of each CMA-ES iteration for cleaning purposes
        """
        pass

    def _finish(self):
        """
        Export current state of CMA
        """
        with open(self.output_dir.cma_pickle, "wb") as f:
            f.write(self.es.pickle_dumps())

        np.save(self.output_dir.pa, self.param_history)
        np.save(self.output_dir.en, self.energy_history)
        np.save(self.output_dir.enf, self.energy_history_ft)
        np.save(self.output_dir.score, self.energy_history_score)
        np.save(self.output_dir.cts, self.energy_history_constraints)
        np.save(self.output_dir.prior, self.energy_history_prior)
        np.save(self.output_dir.fv, self.features_history)

        self.history.dump(self.output_dir.history)
        tf.dump_json(self.output_dir.term, self.termination_options)
        self.time_end_ = time.perf_counter()

        with open(self.output_dir.best_results, "w") as f:
            f.write(f"Time: {datetime.now()}\n\n")
            f.write(f"Computation time: {self.time_end_-self.time_start_:.5f}s\n")
            f.write(f"Stop criteria: {self.es.stop()}\n")
            f.write(f"Total iterations: {self.es.countiter}\n\n")
            f.write(f"Overall best dir: {self.best_dirs.best}\n")
            f.write(f"Best iteration:\n{self.best}\n\n")
            f.write(f"Last best dir: {self.best_dirs.last}\n")
            f.write(f"Last iteration:\n{self.last}\n\n")
            f.write(f"Prior:\n{self.prior}\n\n")

        s = None if isinstance(self.best, str) else self.best.fvalue
        log.info(f"Finishing CMA-ES ({self.es.countiter} iterations, fvalue={s})")
        self._finished = True

    def set_opts(self, **kwargs):
        """
        Set pycma options
        """
        self.es.opts.update(kwargs)
        self.dump_options()

    @property
    def options(self):
        """
        Get pycma options
        """
        return self.es.opts

    @options.setter
    def options(self, d: dict):
        """
        Set pycma options
        """
        self.set_opts(**d)

    def zip(self):
        """
        Create a zipped archive of exported data
        """
        fname = self.output_dir.path(f"{self.output_dir.basename}.zip")
        tf.dump_zip(fname, self.output_dir)
        log.info(f"Saved zip to {fname!r}")


log = logging.getLogger(__name__)
